#include "pattern.h"
#include "shift_register.h"
#include "ring_buffer16.h"

#include <avr/io.h>
#include <avr/interrupt.h>

#define SAMPLE_BUFFER_SIZE	202

enum EState {IDLE,SAMPLE, SET, SAMPLE_AND_SET, SERVO, TRIGGERING};

static volatile uint8_t timer_ovf = 0;

static uint16_t trigger_1;
static uint16_t trigger_2;
static uint8_t reset_trigger;
static enum EState		state;
static enum EState		state_after_trigger;

static uint16_t servo_ocrs[8];
static uint8_t servo_index;

static uint8_t n_samples;

static RingBuffer16Data 	out_data[SAMPLE_BUFFER_SIZE];
static RingBuffer16Data		in_data[SAMPLE_BUFFER_SIZE];

static struct SRingBuffer16	in_buffer;
static struct SRingBuffer16	out_buffer;

ISR(TIMER1_COMPB_vect)
{
	ShiftRegister_SetOutput(0);
}

ISR(TIMER1_COMPA_vect)
{
	timer_ovf++;
}

void Pattern_Init()
{
	RingBuffer16_Init(&out_buffer, out_data, SAMPLE_BUFFER_SIZE );
	RingBuffer16_Init(&in_buffer, in_data, SAMPLE_BUFFER_SIZE );
	state = IDLE;

	Pattern_SetTimerParameter( (1 << CS11), 0x0733 ); //1ms intervall
}

void Pattern_SetTrigger(uint16_t t1, uint16_t t2)
{
	trigger_1 = t1;
	trigger_2 = t2;
}

void 	Pattern_SetTimerParameter(uint8_t prescaler, uint16_t compare_value)
{
	prescaler &= ((1<<CS12) | (1<<CS11) | (1<< CS10)); //mask invalid bits

	if(prescaler >= 0x06) //don't set an invalid mode (ext clock)
		return;

	TCCR1B = prescaler | (1<< WGM12);
	OCR1A = compare_value;
}

uint8_t		Pattern_Triggered(void)
{
	static uint16_t input,last_input;

	if(reset_trigger)
	{
		last_input = ShiftRegister_GetInput();
		reset_trigger = 0;
	}
	else
	{
		input = ShiftRegister_GetInput();
	}

/*
		t1    	t2
		0		0		no trigger
		1		0		rising edge
		0		1		falling edge
		1		1		rising and fallign edge
*/

	if(!(trigger_1 | trigger_2) | (trigger_1 & input & !last_input) | (trigger_2 & !input & last_input))
	{
		last_input = input;
		//process the first data sample, that invoked the trigger
		if(state == SAMPLE)
		{
			RingBuffer16_Push(input, &in_buffer);
			n_samples--;
			if(n_samples == 0)
				state = IDLE;
		}
		else if (state == SAMPLE_AND_SET)
		{
			RingBuffer16_Push(input, &in_buffer);

			uint16_t out = RingBuffer16_Pop(&out_buffer);
			ShiftRegister_SetOutput(out);

			n_samples--;
			if(n_samples == 0)
				state = IDLE;
		}
		return 1;
	}

	last_input = input;
	return 0;
}

static void Pattern_Reset(uint8_t n)
{
	TIMSK &= ~(1<<OCIE1A);
	TIMSK &= ~(1<<OCIE1B);
	RingBuffer16_Clear(&in_buffer);
	RingBuffer16_Clear(&out_buffer);
	state = IDLE;
	n_samples = n;
	timer_ovf = 0;
	reset_trigger = 1;
}

void Pattern_SetOutput(uint16_t out)
{
	Pattern_Reset(0);
	ShiftRegister_SetOutput(out);
}

uint16_t Pattern_GetInput(void)
{
	Pattern_Reset(0);
	return ShiftRegister_GetInput();
}

void Pattern_Sample(uint8_t n)
{
	Pattern_Reset(n);

	if( trigger_1 || trigger_2 )
	{
		state = TRIGGERING;
		state_after_trigger = SAMPLE;
	}
	else
		state = SAMPLE;

	TIMSK |= (1<<OCIE1A);
}

void Pattern_Pattern(uint8_t n)
{
	Pattern_Reset(n);

	if( trigger_1 || trigger_2 )
	{
		state = TRIGGERING;
		state_after_trigger = SET;
	}
	else
		state = SET;

	// pattern is started, when enough "output data" was written
}

void Pattern_LogicOp(uint8_t n)
{
	Pattern_Reset(n);

	if( trigger_1 || trigger_2 )
	{
		state = TRIGGERING;
		state_after_trigger = SAMPLE_AND_SET;
	}
	else
		state = SAMPLE_AND_SET;

	// pattern is started, when enough "output data" was written
}


void Pattern_AddOutputData(uint16_t	out_data)
{
	RingBuffer16_Push(out_data, &out_buffer);

	if(RingBuffer16_Size( &out_buffer ) == n_samples)
	{
		TIMSK |= (1<<OCIE1A);
	}
}

void Pattern_SetServo(uint16_t ocrs[8])
{
	if(state != SERVO)
	{
		servo_index = 0;
		state = SERVO;
		timer_ovf = 0;
		//set correct timer frequency: Prescaler 8, 2.5 ms
		Pattern_SetTimerParameter((1<<CS11),4608);
		TIMSK |= (1<<OCIE1A) | (1<<OCIE1B);
	}

	for(uint8_t n = 0; n < 8;n++)
	{
		servo_ocrs[n] = ocrs[n];
	}
}

uint8_t Pattern_GetInputData(uint16_t *data)
{
	if( RingBuffer16_Size( &in_buffer ) )
	{
		*data = RingBuffer16_Pop(&in_buffer);
		return 1;
	}
	return 0;
}

void Pattern_Process(void)
{
	if(timer_ovf == 0)
		return;

	timer_ovf--;

	switch(state)
	{
		case IDLE:
		{

		}break;
		case SAMPLE:
		{
			uint16_t in = ShiftRegister_GetInput();
			RingBuffer16_Push(in, &in_buffer);

			n_samples--;
			if(n_samples == 0)
			{
				state = IDLE;
				TIMSK &= ~(1<<OCIE1A);
			}
		}break;
		case SET:
		{
			uint16_t out = RingBuffer16_Pop(&out_buffer);
			ShiftRegister_SetOutput(out);

			n_samples--;
			if(n_samples == 0)
			{
				state = IDLE;
				TIMSK &= ~(1<<OCIE1A);
			}
		}break;
		case SAMPLE_AND_SET:
		{
			uint16_t in = ShiftRegister_GetInput();
			RingBuffer16_Push(in, &in_buffer);

			uint16_t out = RingBuffer16_Pop(&out_buffer);
			ShiftRegister_SetOutput(out);

			n_samples--;
			if(n_samples == 0)
			{
				state = IDLE;
				TIMSK &= ~(1<<OCIE1A);
			}
		}break;
		case TRIGGERING:
		{
			if(Pattern_Triggered())
				state = state_after_trigger;
		}break;
		case SERVO:
		{
			ShiftRegister_SetOutput( 1<< servo_index );
			TCNT1 = 0;
			OCR1B = servo_ocrs[servo_index];
			servo_index++;
			servo_index &= 0x07;
		}break;
		default:break;
	};
}


